import pickle
import threading
import contextlib
import re

import decorator

import kazoo.client
import kazoo.exceptions

from contextlog import get_logger

from ...core.optconf import Option
from ...core.optconf.converters import as_string_or_none
from ...core.backends import ConnectionError


# =====
class NoNodeError(Exception):
    pass


class NodeExistsError(Exception):
    pass


class NotEmptyError(Exception):
    pass


class EmptyValue:  # pylint: disable=no-init
    def __new__(cls):
        raise RuntimeError("Use a class rather than an object of class")


# =====
def join(base, *items):
    is_root = (base.startswith("/") or len(base) == 0)
    path = base + "/" + "/".join(items)
    path = "/".join(filter(None, path.split("/")))
    if is_root:
        path = "/" + path
    return path


# =====
def _encode_value(value):
    if value is EmptyValue:
        return b""
    else:
        return pickle.dumps(value)


def _decode_value(value):
    assert isinstance(value, bytes), "Invalid ZK output value type: {}".format(type(value))
    if len(value) == 0:
        return EmptyValue
    else:
        return pickle.loads(value)


def _catch_zk(method):
    def wrap(method, *args, **kwargs):
        try:
            return method(*args, **kwargs)
        except kazoo.exceptions.NoNodeError:
            raise NoNodeError
        except kazoo.exceptions.NodeExistsError:
            raise NodeExistsError
        except kazoo.exceptions.NotEmptyError:
            raise NotEmptyError
    return decorator.decorator(wrap, method)


def _catch_zk_conn(method):
    def wrap(method, *args, **kwargs):
        try:
            return method(*args, **kwargs)
        except kazoo.exceptions.ConnectionLoss:
            raise ConnectionError
    return decorator.decorator(wrap, method)


# ====
class Client:
    """
        This interface provides a simple low-level abstraction for ZooKeeper (using kazoo),
        with a friendly interface and the logged write operations.
    """

    def __init__(self, nodes, timeout, start_timeout, start_retries, randomize_hosts, chroot):
        assert isinstance(nodes, (list, tuple))
        for node in nodes:
            assert re.match(r"[^:]+:\d+", node) is not None, "zookeeper node should has format host:port"
        self._hosts = ",".join(nodes)
        self._timeout = timeout
        self._start_timeout = start_timeout
        self._start_retries = start_retries
        self._randomize_hosts = randomize_hosts
        self._chroot = chroot
        self.zk = None

    @classmethod
    def get_options(cls):
        return {
            "nodes": Option(default=["localhost:2181"], help="List of hosts to connect (in host:port format)"),
            "timeout": Option(default=10.0, help="The longest to wait for a Zookeeper connection"),
            "start_timeout": Option(default=10.0, help="Timeout of the initial connection"),
            "start_retries": Option(default=1, type=int, help="The number of attempts the initial "
                                                              "connection to ZooKeeper (0=infinite)"),
            "randomize_hosts": Option(default=True, help="Randomize host selection"),
            "chroot": Option(default=None, type=as_string_or_none, help="Use specified node as root "
                                                                        "(it must be created manually)"),
        }

    @contextlib.contextmanager
    def connected(self):
        self.open()
        try:
            yield self
        finally:
            self.close()

    def _ensure_chroot(self):
        with Client(
            nodes=self._hosts.split(","),
            timeout=self._timeout,
            start_timeout=self._start_timeout,
            start_retries=self._start_retries,
            randomize_hosts=self._randomize_hosts,
            chroot=None,
        ).connected() as client:
            try:
                with client.make_write_request("ensure_chroot()") as request:
                    request.create(self._chroot, recursive=True)
            except NodeExistsError:
                pass

    @_catch_zk_conn
    def open(self):
        if self._chroot is not None:
            self._ensure_chroot()

        self.zk = kazoo.client.KazooClient(
            hosts=self._hosts,
            timeout=self._timeout,
            randomize_hosts=self._randomize_hosts,
            command_retry={"max_delay": min(self._start_timeout, self._start_retries)},
        )
        if self._chroot is not None:
            self.zk.chroot = self._chroot

        logger = get_logger()

        start_retries = self._start_retries
        while True:
            remaining = ("inf" if start_retries is None else start_retries)
            logger.debug("Trying to connect to ZK, attempts remaining: %s (timeout: %d)",
                         remaining, self._start_timeout, hosts=self._hosts)
            try:
                self.zk.start(timeout=self._start_timeout)
                break
            except Exception:
                logger.exception("Can't connect to ZK in this time")
                if start_retries is not None:
                    if start_retries > 0:
                        start_retries -= 1
                    else:
                        raise
        logger.debug("Started ZK client", hosts=self._hosts)

    def close(self):
        assert self.zk is not None, "Can't close() not opened client"
        self.zk.stop()
        self.zk.close()
        self.zk = None
        get_logger().debug("ZK client has been closed", hosts=self._hosts)

    def is_opened(self):
        return (self.zk is not None)

    def is_alive(self):
        return (False if self.zk is None else self.zk.connected)

    # ===

    def get_server_info(self):
        # http://zookeeper.apache.org/doc/current/zookeeperAdmin.html#sc_zkCommands
        return {
            "envi": self._get_info_envi(),
            "mntr": self._get_info_mntr(),
        }

    @_catch_zk_conn
    def _get_info_envi(self):
        # $ echo envi | netcat 127.0.0.1 2181
        # Environment:
        # zookeeper.version=3.4.6-1569965, built on 02/20/2014 09:09 GMT
        # host.name=localhost.localdomain
        # ...
        return dict(
            item.split("=", 1)
            for item in self.zk.command(b"envi").split("\n")[1:-1]
        )

    @_catch_zk_conn
    def _get_info_mntr(self):
        # $ echo mntr | netcat 127.0.0.1 2181
        # zk_version      3.4.6-1569965, built on 02/20/2014 09:09 GMT
        # zk_avg_latency  0
        # ...
        return dict(
            item.split("\t", 1)
            for item in self.zk.command(b"mntr").split("\n")[:-1]
        )

    # ===

    @_catch_zk
    @_catch_zk_conn
    def get_children(self, path):
        return self.zk.get_children(path)

    @_catch_zk
    @_catch_zk_conn
    def get_children_count(self, path):
        stat = self.zk.retry(self.zk.get, path)[1]
        return stat.children_count

    @_catch_zk
    @_catch_zk_conn
    def exists(self, path, watch=None):
        return (self.zk.exists(path, watch=watch) is not None)

    @_catch_zk_conn
    def get(self, path, default=EmptyValue):
        try:
            return _decode_value(self.zk.get(path)[0])
        except kazoo.exceptions.NoNodeError:
            if default is EmptyValue:
                raise NoNodeError
            return default

    def make_write_request(self, comment="<unnamed>"):
        return _WriteRequest(self, comment)

    # ===

    def get_lock(self, path, comment="<unnamed>"):
        return _Lock(self, path, comment)


class _WriteRequest:
    """
        This class is used to perform write operations in a single context.
        For several operations, the transaction is automatically used.
    """

    def __init__(self, client, comment):
        self._client = client
        self._comment = comment
        self._ops = []

    def create(self, path, value=EmptyValue, ephemeral=False, sequence=False, recursive=False):
        kwargs = {
            "path": path,
            "value": _encode_value(value),
            "ephemeral": ephemeral,
            "sequence": sequence,
        }
        if recursive:
            kwargs["makepath"] = True  # XXX: Only for a single operation!
        self._ops.append(("create", kwargs))

    def set(self, path, value=EmptyValue):
        self._ops.append(("set", {
            "path": path,
            "value": _encode_value(value),
        }))

    def delete(self, path, recursive=False):
        kwargs = {"path": path}
        if recursive:
            kwargs["recursive"] = True  # XXX: Only for a single operation!
        self._ops.append(("delete", kwargs))

    def __enter__(self):
        get_logger().debug("Created write-request", comment=self._comment)
        return self

    @_catch_zk  # for "len(self._ops) == 1"
    @_catch_zk_conn
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value is not None:
            raise exc_value
        assert len(self._ops) > 0, "_WriteRequest() does not contain operations"
        if len(self._ops) == 1:
            (op_name, kwargs) = self._ops[0]
            getattr(self._client.zk, op_name)(**kwargs)
        else:
            trans = self._client.zk.transaction()
            for (op_name, kwargs) in self._ops:
                if op_name == "set":
                    op_name = "set_data"
                getattr(trans, op_name)(**kwargs)
            results = trans.commit()
            need_err = False
            for result in reversed(results):
                if isinstance(result, kazoo.exceptions.RuntimeInconsistency):
                    need_err = True
                    continue
                elif isinstance(result, Exception):
                    raise result
            assert not need_err, "No other exceptions, but runtime is inconsistent: {}".format(results)
        get_logger().debug("Completed write-request", comment=self._comment)


# =====
class _Lock:
    """
        Easy lock primitive with no queue of those who try to acquire it.
        It can be acquired and released in the transaction.
    """

    def __init__(self, client, path, comment):
        self._client = client
        self._path = path
        self._comment = comment

    @_catch_zk_conn
    def is_locked(self):
        return (self._client.zk.exists(self._path) is not None)

    def acquire(self, request, value=EmptyValue):
        request.create(self._path, value, ephemeral=True)

    def release(self, request):
        request.delete(self._path)

    @_catch_zk_conn
    def __enter__(self):
        get_logger().debug("Acquiring lock", comment=self._comment)
        while not self._try_acquire():
            wait = threading.Event()
            if self._client.zk.exists(self._path, watch=lambda _: wait.set()) is not None:
                wait.wait()

    @_catch_zk_conn
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self._client.zk.delete(self._path)
        except kazoo.exceptions.NoNodeError:
            pass
        get_logger().debug("Released lock", comment=self._comment)

    @_catch_zk_conn
    def _try_acquire(self):
        try:
            self._client.zk.create(self._path, ephemeral=True)
            return True
        except kazoo.exceptions.NodeExistsError:
            return False



import sys
import argparse
import pkgutil
import threading
import socket
import logging
import logging.config
import time
import abc

import yaml

import pygments.lexers.data
import pygments.formatters

from contextlog import (
    get_logger,
    patch_logging,
    patch_threading,
)

from .. import (
    tools,
    backends,
    backdoor,
)

from ..optconf import (
    build_raw_from_options,
    make_config,
    Option,
)

from ..optconf.dumper import make_config_dump
from ..optconf.loader import load_file as load_yaml_file

from ..optconf.converters import (
    as_string_or_none,
    as_int_or_none,
)


# =====
_config = None


def get_config():
    return _config


def init(name, description, args=None, raw_config=None):
    global _config
    assert _config is None, "init() has already been called"

    args_parser = argparse.ArgumentParser(prog=name, description=description)
    args_parser.add_argument("-v", "--version", action="version", version=tools.get_version())
    args_parser.add_argument("-c", "--config", dest="config_file_path", default=None, metavar="<file>")
    args_parser.add_argument("-l", "--level", dest="log_level", default=None)
    args_parser.add_argument("-m", "--dump-config", dest="dump_config", action="store_true")
    args_parser.add_argument("-o", "--set-options", dest="set_options", default=[], nargs="+")
    options = args_parser.parse_args(args)

    # Load configs
    raw_config = (raw_config or {})
    if options.config_file_path is not None:
        raw_config = load_yaml_file(options.config_file_path)
    _merge_dicts(raw_config, build_raw_from_options(options.set_options))
    scheme = _get_config_scheme()
    config = make_config(raw_config, scheme)

    # Configure logging
    patch_logging()
    patch_threading()
    logging.setLogRecordFactory(_ClusterLogRecord)
    logging.captureWarnings(True)
    logging_config = raw_config.get("logging")
    if logging_config is None:
        logging_config = yaml.load(pkgutil.get_data(__name__, "configs/logging.yaml"))
    if options.log_level is not None:
        logging_config.setdefault("root", {})
        logging_config["root"]["level"] = _valid_log_level(options.log_level)
    logging.config.dictConfig(logging_config)

    # Update scheme for backend opts
    backend_scheme = backends.get_backend_class(config.core.backend).get_options()
    _merge_dicts(scheme, {"backend": backend_scheme})
    config = make_config(raw_config, scheme)

    # Provide global configuration
    _config = make_config(raw_config, scheme)

    # Print config dump and exit
    if options.dump_config:
        dump = make_config_dump(config)
        if sys.stdout.isatty():
            dump = pygments.highlight(
                dump,
                pygments.lexers.data.YamlLexer(),
                pygments.formatters.TerminalFormatter(bg="dark"),
            )
        print(dump)
        sys.exit(0)

    return _config


_fqdn_cached = 0


class _ClusterLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # XXX: No locks!
        # http://bugs.python.org/issue6721
        global _fqdn_cached
        stamp = int(time.time() // 60)
        if _fqdn_cached != stamp:
            self.fqdn = socket.getfqdn()
            _fqdn_cached = stamp
        self.node = tools.get_node_name()  # Nodename from uname


class Application(metaclass=abc.ABCMeta):
    def __init__(self, app_name, config):
        self._app_name = app_name
        self._app_config = config[app_name]
        self._config = config
        self._stop_event = threading.Event()
        self._respawns = 0
        self._next_write_state_at = 0

    def stop(self):
        self._stop_event.set()

    def dump_app_state(self, backend, app_state):
        if self._next_write_state_at <= time.time():
            app_state = app_state.copy()
            app_state["respawns"] = self._respawns
            backend.system_apps_state.set_state(self._app_name, app_state)
            self._next_write_state_at = time.time() + self._app_config.state_every

    def get_backend_object(self):
        return backends.get_backend_class(self._config.core.backend)(**self._config.backend)

    ###

    def run(self):
        logger = get_logger(app=self._app_name)  # App-level context
        if self._config.backdoor.enabled:
            backdoor.start(self._config.backdoor.port)
        self._respawns = 0
        while not self._stop_event.is_set():
            if self._app_config.max_fails is not None and self._respawns >= self._app_config.max_fails + 1:
                logger.critical("Reached the respawn maximum, exiting...")
                return -1
            try:
                logger.critical("Ready to work")
                self.process()
            except KeyboardInterrupt:
                logger.critical("Received Ctrl+C, exiting...")
                return 0
            except Exception:
                logger.critical("Error in main loop, respawn...", exc_info=True)
                logger.warning("Sleeping %f seconds...", self._app_config.fail_sleep)
                time.sleep(self._app_config.fail_sleep)
                self._respawns += 1
        return 0

    @abc.abstractmethod
    def process(self):
        raise NotImplementedError


# =====
def _valid_log_level(arg):
    try:
        return int(arg)
    except ValueError:
        return logging._nameToLevel[arg.upper()]  # pylint: disable=protected-access


def _merge_dicts(dest, src, path=None):
    if path is None:
        path = []
    for key in src:
        if key in dest:
            if isinstance(dest[key], dict) and isinstance(src[key], dict):
                _merge_dicts(dest[key], src[key], list(path) + [str(key)])
                continue
        dest[key] = src[key]


def _get_config_scheme():
    scheme = {
        "core": {
            "node_name": Option(default=None, type=as_string_or_none, help="Short node name (like uname -n)"),
            "backend": Option(default="zookeeper", help="Backend plugin"),
            "scripts_dir": Option(default="scripts", help="Path to scripts root"),
        },

        "backdoor": {
            "enabled": Option(default=False, help="Enable telnet-based backdoor to Python process"),
            "port": Option(default=2200, help="Backdoor port"),
        },

        "api": {
            "backend_connections": Option(default=1, help="Maximum number of backend connections"),
            "delete_timeout": Option(default=15, help="Timeout for stop/delete operation"),
            "gunicorn": Option(default={}, help="Gunicorn options (workers, max_requests, etc.) "
                                                "exclude entrypoint-specific (like errorlog, accesslog). "
                                                "See http://docs.gunicorn.org/en/latest/settings.html"),
        },

        "worker": {
            "max_jobs_sleep": Option(default=1, help="If we have reached the maximum concurrent jobs - "
                                                     "the process goes to sleep (seconds)"),
            "max_jobs": Option(default=None, type=as_int_or_none, help="The maximum number of job processes"),
            "job_delay": Option(default=0.1, help="Sleep between start of jobs"),
            "wait_slowpokes": Option(default=30.0, help="Wait slow jobs before kill"),
        },

        "collector": {},
    }
    for app in ("worker", "collector"):
        scheme[app].update({
            "max_fails": Option(default=None, type=as_int_or_none, help="Number of failures after which "
                                                                        "the program terminates"),
            "fail_sleep": Option(default=5, help="If processing fails, sleep for awhile and restart (seconds)"),
            "empty_sleep": Option(default=1, help="Interval after which process will sleep when "
                                                  "there are no jobs (seconds)"),
            "state_every": Option(default=5, help="Dump appstate to backend every this time (seconds)"),
        })
    return scheme



import logging

import flask
import flask_api.app

import gunicorn.app.base
import gunicorn.glogging

from contextlog import get_logger

from .. import (
    backends,
    impscripts,
    api,
    backdoor,
)

from ..api.scripts import (
    ExposedScriptsResource,
    ScriptsHeadResource,
)

from ..api.jobs import (
    JobsResource,
    JobControlResource,
)

from ..api.cas import (
    CasRootResource,
    CasPathResource,
)

from ..api.system import (
    StateResource,
    InfoResource,
    ConfigResource,
)

from . import init


# =====
class _Api(flask_api.app.FlaskAPI):
    """
        Versioned REST API.
    """

    def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
        # Yes, __init__ not from a base class
        flask.Flask.__init__(self, *args, **kwargs)  # pylint: disable=non-parent-init-called
        self.api_settings = flask_api.app.APISettings(self.config)
        self.jinja_env.filters["urlize_quoted_links"] = flask_api.app.urlize_quoted_links
        self.register_blueprint(flask.Blueprint(
            name="flask-api",
            import_name="flask_api.app",
            url_prefix="/flask-api",
            static_folder="static",
        ))
        self.register_blueprint(flask.Blueprint(
            name="powny-api",
            import_name="powny.core.api",
            template_folder="templates",
        ))
        self._resources = {}
        self.add_url_rule("/", "API's list", self._get_apis)
        self._count = 0

    def add_url_resource(self, version, url_rule, resource):
        def handler(**kwargs):
            return resource.handler(**kwargs)
        handler.__doc__ = resource.docstring
        self.add_url_rule(
            url_rule,
            str(resource.__class__.__name__),
            handler,
            methods=resource.methods,
        )
        self._resources.setdefault(version, [])
        self._resources[version].append(resource)
        self._count += 1

    def _get_apis(self):
        """ View available API's """
        return {
            version: [
                {"name": resource.name, "url": api.get_url_for(resource)}
                for resource in resources
                if not resource.dynamic
            ]
            for (version, resources) in self._resources.items()
        }


# =====
def make_app(config):
    """
        Use this functions without arguments to create UWSGI app.
    """

    if config.backdoor.enabled:
        backdoor.start(config.backdoor.port)

    pool = backends.Pool(
        size=config.api.backend_connections,
        backend_name=config.core.backend,
        backend_opts=config.backend,
    )

    loader = impscripts.Loader(config.core.scripts_dir)

    app = _Api(__name__)
    app.add_url_resource("v1", "/v1/scripts/exposed", ExposedScriptsResource(pool, loader))
    app.add_url_resource("v1", "/v1/scripts/head", ScriptsHeadResource(pool))
    app.add_url_resource("v1", "/v1/jobs", JobsResource(pool, loader))
    app.add_url_resource("v1", "/v1/jobs/<job_id>", JobControlResource(pool, config.api.delete_timeout))
    app.add_url_resource("v1", "/v1/cas", CasRootResource(pool))
    app.add_url_resource("v1", "/v1/cas/<path:path>", CasPathResource(pool))
    app.add_url_resource("v1", "/v1/system/state", StateResource(pool))
    app.add_url_resource("v1", "/v1/system/info", InfoResource(pool))
    app.add_url_resource("v1", "/v1/system/config", ConfigResource(config))

    return app


def run(args=None, config=None):
    get_logger(app="api")  # App-level context
    if config is None:
        config = init(__name__, "Powny HTTP API", args)
    _Unicorn(config).run()


# =====
class _Unicorn(gunicorn.app.base.BaseApplication):
    def __init__(self, config):
        self._config = config
        super().__init__()

    def init(self, parser, opts, args):
        pass  # Makes pylint happy

    def load_config(self):
        for (option, value) in self._config.api.gunicorn.items():
            self.cfg.set(option, value)
        self.cfg.set("logger_class", _UnicornLogger)
        self.cfg.set("accesslog", "-")  # For working accesslog (see gunicorn/glogging.py:270)

    def load(self):
        app = make_app(self._config)
        get_logger().critical("Ready to work on %s", self.cfg.bind)
        return app


class _UnicornLogger(gunicorn.glogging.Logger):
    def __init__(self, cfg):  # pylint: disable=super-init-not-called
        self.error_log = logging.getLogger("gunicorn.error")
        self.access_log = logging.getLogger("gunicorn.access")
        self.cfg = cfg

    def setup(self, _):
        pass  # Don't configure logging using the gunicorn options



import re

from flask import request

from . import (
    get_exposed,
    Resource,
    ApiError,
)


# =====
class ExposedScriptsResource(Resource):
    name = "Information about scripts"
    docstring = """
        GET  -- Returns a current version (head) of the scripts in format:

                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {
                        "head":    "<HEAD>"
                        "errors":  {"<path.to.module>": "<Traceback>", ...}
                        "exposed": ["<path.to.function>", ...]|null,
                    },
                }
                # =====

                @head    -- Current version of the scripts. Null if the version has not yet been set.
                @errors  -- Errors that occurred while loading the specified modules (null if global error occurs).
                @exposed -- List of functions that can be called directly by name (null if global error occurs).

                Possible errors (with status=="error"):
                    503 -- Non-existant HEAD for scripts.
    """

    def __init__(self, pool, loader):
        self._pool = pool
        self._loader = loader

    def process_request(self):
        backend = self._pool.get_backend()
        try:
            if request.method == "GET":
                (head, exposed, errors, exc) = get_exposed(backend, self._loader)
                if exc is None:  # No errors
                    # exposed is None if HEAD is not configured
                    exposed_names = (list(exposed) if exposed is not None else None)
                    return ({"head": head, "exposed": exposed_names, "errors": errors}, "The scripts of current HEAD")
                else:
                    raise ApiError(503, exc, {"head": head, "exposed": None, "errors": None})
        finally:
            self._pool.retrieve_backend(backend)


class ScriptsHeadResource(Resource):
    name = "Operations with scripts HEAD"
    methods = ("GET", "POST")
    docstring = """
        GET  -- Returns a current version (head) of the scripts in format:

                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {"head": "<HEAD>"|null},
                }
                # =====

                @head -- Current version of the scripts. Null if the version has not yet been set.

        POST -- Takes a version (head) in the format: {"head": "<HEAD>"} and applies it.

                Return value:
                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {"head": "<HEAD>"},
                }
                # =====

                Possible POST errors (with status=="error"):
                    400 -- Invalid HEAD (not a hex string).
    """

    def __init__(self, pool):
        self._pool = pool

    def process_request(self):
        backend = self._pool.get_backend()
        try:
            if request.method == "GET":
                return self._request_get(backend)
            elif request.method == "POST":
                return self._request_post(backend)
        finally:
            self._pool.retrieve_backend(backend)

    def _request_post(self, backend):
        head = str((request.data or {}).get("head")).strip()
        if re.match(r"^[0-9a-fA-F]+$", head) is None:
            raise ApiError(400, "The HEAD must be a hex string", {"head": head})
        backend.scripts.set_head(head)
        return ({"head": head}, "The HEAD has been updated")

    def _request_get(self, backend):
        return ({"head": backend.scripts.get_head()}, "Current HEAD")



# flake8: noqa


from .tools import (
    get_version,
    get_user_agent,
    get_node_name,
)

from .apps import get_config

from .impscripts import expose

from .context import (
    SuicideError,
    suicide,
    in_context,
    get_context,
    get_job_id,
    get_extra,
    get_cas_storage,
    save_job_state,
)

__version__ = get_version()



import sys
import os
import signal
import multiprocessing
import threading
import setproctitle
import logging
import errno
import time

from contextlog import get_logger

from .. import context

from . import (
    init,
    Application,
)


# =====
_stop = None


def run(args=None, config=None):
    if config is None:
        config = init(__name__, "Powny Worker", args)
    app = _Worker(config)
    global _stop
    _stop = app.stop
    return abs(app.run())


# =====
class _Worker(Application):
    """
        This application performs the jobs. Each job runs in a separate process and
        has its own connection to the backend. It automatically saves the state of
        the job. Worker only ensures that the process must be stopped upon request.
    """

    def __init__(self, config):
        Application.__init__(self, "worker", config)
        self._jobs_limit = None
        self._procs = {}
        self._finished = 0
        self._not_started = 0

    def process(self):
        logger = get_logger()

        with self.get_backend_object().connected() as backend:
            while not self._stop_event.is_set():
                self._set_jobs_limit(backend)
                self._manage_processes(backend)
                self._dump_worker_state(backend)

                if len(self._procs) >= self._jobs_limit:
                    logger.debug("Have reached the maximum concurrent jobs %d, sleeping %f seconds...",
                                 self._jobs_limit, self._app_config.max_jobs_sleep)
                    time.sleep(self._app_config.max_jobs_sleep)

                elif backend.jobs_process.has_awaiting_jobs():
                    if not self._try_start_job():
                        logger.debug("No jobs to start, sleeping for %f seconds...", self._app_config.empty_sleep)
                        time.sleep(self._app_config.empty_sleep)

    def _manage_processes(self, backend):
        for (job_id, proc) in self._procs.copy().items():
            logger = get_logger(job_id=job_id)

            alive = proc.is_alive()
            deleted = backend.jobs_process.is_deleted_job(job_id)

            if not alive or deleted:
                if deleted:
                    logger.info("Terminating job process %d...", proc.pid)
                    try:
                        proc.terminate()
                        logger.info("Terminated")
                    except Exception:
                        logger.exception("Can't terminate process %d; ignored", proc.pid)
                proc.join()
                self._procs.pop(job_id)
                self._finished += 1
                logger.info("Finished job process %d with retcode %d", proc.pid, proc.exitcode)

    def _set_jobs_limit(self, backend):
        if self._app_config.max_jobs is None:
            all_jobs = backend.jobs_control.get_jobs_count()
            workers = max(len(backend.system_apps_state.get_full_state().get("worker", {})), 1)
            jobs_limit = int(all_jobs / workers + 1)
        else:
            jobs_limit = self._app_config.max_jobs

        if self._jobs_limit != jobs_limit:
            get_logger().info("Set new jobs limit: %d", jobs_limit)
            self._jobs_limit = jobs_limit

    def _dump_worker_state(self, backend):
        self.dump_app_state(backend, {
            "active": len(self._procs),
            "processed": self._finished,
            "not_started": self._not_started,
            "jobs_limit": self._jobs_limit,
        })

    def _try_start_job(self):
        logger = get_logger()
        logger.info("Starting candidate process")

        (read_conn, write_conn) = multiprocessing.Pipe(True)
        proc = multiprocessing.Process(
            target=_find_job_and_run,
            kwargs={
                "backend": self.get_backend_object(),
                "scripts_dir": self._config.core.scripts_dir,
                "conns": (read_conn, write_conn),
            },
        )
        proc.start()
        write_conn.close()

        if read_conn.poll(self._app_config.wait_slowpokes):
            job_id = read_conn.recv()
            if job_id is not None:  # Задача нашлась и запущена
                self._procs[job_id] = proc
                return True
            else:  # Задач нет, надо поспать
                proc.join()
        else:  # Слоупок
            logger.error("Detected slowpoke process %d", proc.pid)
            killed = _send_signal(proc, signal.SIGKILL)
            proc.join()
            if killed:
                logger.info("Killed slowpoke job process %d with retcode %d", proc.pid, proc.exitcode)
            else:
                logger.info("Found dead slowpoke job process %d with retcode %d", proc.pid, proc.exitcode)
            self._not_started += 1
        read_conn.close()


def _send_signal(proc, signum):
    try:
        os.kill(proc.pid, signum)
        return True
    except OSError as err:
        if err.errno == errno.ESRCH:
            return False
        raise


def _find_job_and_run(backend, scripts_dir, conns):
    _unlock_logging()
    (read_conn, write_conn) = conns
    read_conn.close()
    sys.dont_write_bytecode = True
    try:
        with backend.connected():
            job = backend.jobs_process.get_job()
            if job is None:
                get_logger().info("Nothing to run; exit")
                write_conn.send(None)
                write_conn.close()
                return

            scripts_path = os.path.join(scripts_dir, job.head)
            sys.path.insert(0, scripts_path)
            logger = get_logger(job_id=job.job_id)
            logger.info("Starting the job")
            _rename_process(job)

            thread = context.JobThread(
                backend=backend,
                job_id=job.job_id,
                state=job.state,
                extra={"head": job.head},
                fatal_internal=(not job.respawn),
            )
            thread.start()
            write_conn.send(job.job_id)
            write_conn.close()
            thread.join()
    except Exception:
        logger.exception("Unhandled exception in subprocess")
        raise


def _rename_process(job):
    setproctitle.setproctitle("[worker: {}] {}".format(  # pylint: disable=no-member
        job.job_id,
        setproctitle.getproctitle(),  # pylint: disable=no-member
    ))


def _unlock_logging():
    # XXX: ULTIMATE EPIC-SIZED CRUTCH!!!111
    # http://bugs.python.org/issue6721
    if logging._lock:  # pylint: disable=protected-access
        logging._lock = threading.RLock()  # pylint: disable=protected-access
    for wr in logging._handlerList:  # pylint: disable=protected-access
        handler = wr()
        if handler and handler.lock:
            handler.lock = threading.RLock()



import os
import contextlib
import tempfile


# =====
@contextlib.contextmanager
def write_file(text):
    try:
        (fd, path) = tempfile.mkstemp()
        os.close(fd)
        with open(path, "w") as yaml_file:
            yaml_file.write(text)
        yield path
    finally:
        os.remove(path)



import uuid

from flask import request

from ..context import dump_call

from ..backends import (
    JobState,
    DeleteTimeoutError,
)

from . import (
    valid_object_name,
    get_exposed,
    get_url_for,
    Resource,
    ApiError,
)


# =====
class JobsResource(Resource):
    name = "Create and view jobs"
    methods = ("GET", "POST")
    docstring = """
        GET  -- Returns a dict of all jobs in the system:
                # =====
                {
                    {
                        "status":   "ok",
                        "message":  "<...>",
                        "<job_id>": {
                            "url": "<http://api/url/to/control/the/job>"},
                        },
                        ...
                }
                # =====

        POST -- Run the specified method. If the argument "method" is specified, will
                be running the specified method (requires the full path from the scripts).
                The arguments passed to method via request body (in dict format).

                Return value:
                # =====
                {
                    "status":  "ok",
                    "message": "<...>",
                    "result":  {
                        "<job_id>": {
                            "method": "<path.to.function>",
                            "url": "<http://api/url/to/control/the/job>"},
                        },
                    },
                }
                # =====

                Possible POST errors (with status=="error"):
                    400 -- Missed required argument; job already exists.
                    404 -- Method not found (for method call).
                    503 -- No HEAD or exposed methods.
    """

    def __init__(self, pool, loader):
        self._pool = pool
        self._loader = loader

    def process_request(self):
        backend = self._pool.get_backend()
        try:
            if request.method == "GET":
                return self._request_get(backend)
            elif request.method == "POST":
                return self._request_post(backend)
        finally:
            self._pool.retrieve_backend(backend)

    def _request_get(self, backend):
        result = {
            job_id: {"url": self._get_job_url(job_id)}  # TODO: Add "method" key
            for job_id in backend.jobs_control.get_jobs_list()
        }
        return (result, ("No jobs" if len(result) == 0 else "The list with all jobs"))

    def _request_post(self, backend):
        (head, exposed, _, _) = get_exposed(backend, self._loader)
        if exposed is None:
            raise ApiError(503, "No HEAD or exposed methods")

        job_id = valid_object_name(request.args.get("job_id", str(uuid.uuid4())))
        method_name = request.args.get("method", None)
        if method_name is None:
            raise ApiError(400, "Required method_name")
        kwargs = dict(request.data or {})
        respawn = (str(request.args.get("respawn", "false")).lower() in ("1", "true", "yes"))

        method = exposed.get(method_name)
        if method is None:
            raise ApiError(404, "Method not found")

        job = JobState(
            job_id=job_id,
            head=head,
            method_name=method_name,
            kwargs=kwargs,
            state=dump_call(method, kwargs),
            respawn=respawn,
        )
        if not backend.jobs_control.add_job(job):
            raise ApiError(400, "Job already exists")

        result = {job_id: {"method": method_name, "url": self._get_job_url(job_id)}}
        return (result, "Method was launched")

    def _get_job_url(self, job_id):
        return get_url_for(JobControlResource, job_id=job_id)


class JobControlResource(Resource):
    name = "View and stop job"
    methods = ("GET", "DELETE")
    dynamic = True
    docstring = """
        GET    -- Returns the job state:
                  # =====
                  {
                      "status":  "ok",
                      "message": "<...>",
                      "result":  {
                          "method":   "<path.to.function>",  # Full method path in the scripts
                          "head":     "<HEAD>",    # HEAD of the scripts for this job
                          "kwargs":   {...},       # Function arguments
                          "respawn":  <bool>,      # Infinite respawn (except unpickle errors)
                          "created":  <str>,       # ISO-8601-like time when the job was created
                          "locked":   <dict|null>, # Job in progress (null if not locked, dict with info otherwise)
                          "deleted":  <str|null>,  # ISO-8601-like time when job was marked to stop and delete
                          "taken":    <str|null>,  # ISO-8601-like time when job was started (taken from queue)
                          "finished": <str|null>,  # ISO-8601-like time when job was finished
                          "stack":    [...],       # Stack snapshot on last checkpoint
                          "retval":   <any|null>,  # Return value if finished and not failed
                          "exc":      <str|null>,  # Text exception if job was failed
                      },
                  }
                  # =====

        DELETE -- Request to immediate stop and remove the job. Waits until the collector does
                  not remove the job.

                  Return value:
                  # =====
                  {
                      "status":  "ok",
                      "message": "<...>",
                      "result": {"deleted": "<job_id>"},
                  }
                  # =====

                  Possible DELETE errors (with status=="error"):
                      503 -- The collector did not have time to remove the job, try again.

        Errors (with status=="error"):
            400 -- Invalid job id.
            404 -- Job not found (or DELETED).
    """

    def __init__(self, pool, delete_timeout):
        self._pool = pool
        self._delete_timeout = delete_timeout

    def process_request(self, job_id):  # pylint: disable=arguments-differ
        job_id = valid_object_name(job_id)
        backend = self._pool.get_backend()
        try:
            if request.method == "GET":
                return self._request_get(backend, job_id)
            elif request.method == "DELETE":
                return self._request_delete(backend, job_id)
        finally:
            self._pool.retrieve_backend(backend)

    def _request_get(self, backend, job_id):
        job_info = backend.jobs_control.get_job_info(job_id)
        if job_info is None:
            raise ApiError(404, "Job not found")
        return (job_info, "Information about the job")

    def _request_delete(self, backend, job_id):
        wait = request.args.get("wait", self._delete_timeout)
        try:
            wait = max(0, int(wait))
        except ValueError:
            raise ApiError(400, "Invalid wait")
        try:
            deleted = backend.jobs_control.delete_job(job_id, timeout=(wait or None))
        except DeleteTimeoutError as err:
            raise ApiError(503, str(err))
        if not deleted:
            raise ApiError(404, "Job not found")
        else:
            return ({"deleted": job_id}, "The job has been removed")



import traceback
import inspect
import pickle
import copy
import threading

from contextlog import get_logger


# =====
class SuicideError(Exception):
    pass


def suicide(msg):
    raise SuicideError(msg)


def in_context():
    thread = threading.current_thread()
    return isinstance(thread, JobThread)


def get_context():
    assert in_context(), "Called not from a job context!"
    return threading.current_thread()


def get_job_id():
    return get_context().get_job_id()  # pylint: disable=maybe-no-member


def get_extra():
    return get_context().get_extra()  # pylint: disable=maybe-no-member


def get_cas_storage():
    return get_context().get_cas_storage()  # pylint: disable=maybe-no-member


def save_job_state():
    return get_context().save()  # pylint: disable=maybe-no-member


# =====
def dump_call(method, kwargs):
    """ Собирает из метода и его аргументов континулет и пиклит его """

    get_logger().debug("Creating a new continulet...")
    import _continuation
    return pickle.dumps(_continuation.continulet(lambda _: method(**kwargs)))


def restore_call(state):
    """ Распикливает состояние (континулет) для запуска """

    get_logger().debug("Restoring the continulet state...")
    import _continuation
    cont = pickle.loads(state)
    assert isinstance(cont, _continuation.continulet), "The unpickled state is a garbage!"
    return cont


class JobThread(threading.Thread):
    """
        JobThread() предназначен для запуска ранее запикленной в континулет функции (с помощью dump_call()).
        Внутри потока континулет распикливается и исполняется, сохраняя свое состояние в переданный бекенд
        функциями backend.jobs_process.save_job_state() и backend.jobs_process.done_job().
        Поток нужен не только для исполнения континулета, но и для того, чтобы изнутри континулета можно было
        получить метаданные текущей задачи, используя threading.current_thread().
    """

    def __init__(self, backend, job_id, state, extra, fatal_internal,  # pylint: disable=unused-argument
                 __unpickle=False):

        threading.Thread.__init__(self, name="JobThread::" + job_id)
        self._backend = backend
        self._job_id = job_id
        self._state = state
        self._extra = extra
        self._fatal_internal = fatal_internal
        self._cont = None
        self._log_context = get_logger().get_context()  # Proxy context into the continulet

    def __getstate__(self):
        # Шаг 1. Запикливаемся, как строка с айдишником задачи. Он нужен только для того, чтобы
        #        проверить, что потом мы распиклились в новом контексте, но с тем же айдишником.
        #        При распикливании новый объект не создается, а возвращается ссылка на текущий
        #        контекст, чтобы одну и ту же задачу не представляло два разных объекта (хоть и
        #        с одинаковым стейтом).
        return self._job_id

    def __getnewargs__(self):
        # Шаг 2. Распикливаемся, передавая в __new__ аргумент о том, что мы именно распикливаемся,
        #        а не создаем новый контекст.
        #        https://docs.python.org/3.2/library/pickle.html#pickle.object.__getnewargs__
        return ((None,) * 5) + (True,)  # Unpickle as current context

    def __new__(cls, backend, job_id, state, extra, fatal_internal, __unpickle=False):
        if __unpickle:
            # Шаг 3. При распикливании, вместо создания нового объекта, возвращаем ссылку на текущий
            #        контекст, предполагая, что он и является контекстом той задачи, в которой
            #        произошло распикливание.
            context = get_context()
            return context  # Return the current context instead of the new object
        else:
            return super(JobThread, cls).__new__(cls, backend, job_id, state, extra, fatal_internal)

    def __setstate__(self, job_id):
        # Шаг 4. После распикливания, проверяем, что мы распиклились в правильном контексте. Айдишник
        #        задачи текущего контекста должен совпадать с айдишником объекта и с тем айдишником,
        #        контекст которого хотели распиклить.
        assert get_context().get_job_id() == self._job_id == job_id  # pylint: disable=maybe-no-member

    ###

    def get_job_id(self):
        return self._job_id

    def get_extra(self):
        return copy.deepcopy(self._extra)

    def get_cas_storage(self):
        return self._backend.cas_storage

    def save(self):
        stack = traceback.extract_stack(inspect.currentframe())
        self._cont.switch(stack)

    ###

    def run(self):
        logger = get_logger(**self._log_context)

        logger.debug("Initializing context...")
        try:
            self._cont = restore_call(self._state)
        except Exception:
            logger.exception("Context initialization has failed")
            self._backend.jobs_process.done_job(
                job_id=self._job_id,
                retval=None,
                exc=traceback.format_exc(),
            )
            raise

        logger.debug("Activation...")
        while self._cont.is_pending():
            try:
                logger.debug("Entering continulet...")
                stack_or_retval = self._cont.switch()
                logger.debug("Exited from continulet")
                if self._cont.is_pending():  # In progress
                    self._backend.jobs_process.save_job_state(
                        job_id=self._job_id,
                        state=pickle.dumps(self._cont),
                        stack=stack_or_retval,
                    )
                else:  # Done
                    self._backend.jobs_process.done_job(
                        job_id=self._job_id,
                        retval=stack_or_retval,
                        exc=None,
                    )
            except Exception as err:
                if isinstance(err, SuicideError):
                    logger.warning("Suicide; fatal_internal=%s", self._fatal_internal)
                else:
                    logger.exception("Unhandled step error; fatal_internal=%s", self._fatal_internal)
                # self._cont.switch() switches the stack, so we will see a valid exception, up to this place
                # in the script. sys.exc_info() return a raw exception data. Some of them can't be pickled, for
                # example, traceback-object. For those who use the API, easier to read the text messages.
                # traceback.format_exc() simply converts data from sys.exc_info() into a string.
                if self._fatal_internal:
                    self._backend.jobs_process.done_job(
                        job_id=self._job_id,
                        retval=None,
                        exc=traceback.format_exc(),
                    )
                logger.debug("Context failed")
                return
        logger.debug("Context finished")



import os
import yaml
import yaml.loader


# =====
def load_file(file_path):
    with open(file_path) as yaml_file:
        try:
            return yaml.load(yaml_file, _YamlLoader)
        except Exception:
            # Reraise internal exception as standard ValueError and show the incorrect file
            raise ValueError("Incorrect YAML syntax in file '{}'".format(file_path))


# =====
class _YamlLoader(yaml.loader.Loader):  # pylint: disable=too-many-ancestors
    def __init__(self, yaml_file):
        yaml.loader.Loader.__init__(self, yaml_file)
        self._root = os.path.dirname(yaml_file.name)

    def include(self, node):
        # Logger which supports include-files
        file_path = os.path.join(self._root, self.construct_scalar(node))  # pylint: disable=no-member
        return load_file(file_path)
_YamlLoader.add_constructor("!include", _YamlLoader.include)  # pylint: disable=no-member



import platform
import datetime
import calendar
import time

import dateutil.parser


# =====
def get_node_name():
    from .apps import get_config
    node_name = (get_config() or {}).get("core", {}).get("node_name")
    return (node_name or platform.uname()[1])


def get_version():
    return "4.2.0"


def get_user_agent():
    return "Powny/{}".format(get_version())


# =====
def make_isotime(unix=None):  # ISO 8601
    if unix is None:
        unix = time.time()
    return datetime.datetime.utcfromtimestamp(unix).strftime("%Y-%m-%d %H:%M:%S.%fZ")


def from_isotime(line):
    dt = dateutil.parser.parse(line)
    return calendar.timegm(dt.utctimetuple()) + dt.microsecond / 10 ** 6  # pylint: disable=maybe-no-member



import time

from contextlog import get_logger

from . import (
    init,
    Application,
)


# =====
_stop = None


def run(args=None, config=None):
    if config is None:
        config = init(__name__, "Powny Garbage Collector", args)
    app = _Collector(config)
    global _stop
    _stop = app.stop
    return abs(app.run())


# =====
class _Collector(Application):
    """
        This application provides cleaning storage of data left after the execution
        of the jobs. Incomplete jobs (due to the failure) are returned to the input
        queue. The completed jobs after expiration of the lifetime are deleted.
    """

    def __init__(self, config):
        Application.__init__(self, "collector", config)
        self._processed = 0

    def process(self):
        logger = get_logger()
        with self.get_backend_object().connected() as backend:
            sleep_mode = False
            while not self._stop_event.is_set():
                sleep_mode = (not self._gc_jobs(backend))  # Separate function for a different log context
                if not sleep_mode:
                    logger.debug("No jobs in list, entering to sleep mode with interval  %f seconds...",
                                 self._app_config.empty_sleep)
                sleep_mode = True
                time.sleep(self._app_config.empty_sleep)

    def _gc_jobs(self, backend):
        processed = 0
        self._dump_collector_state(backend)
        for (job_id, done) in backend.jobs_gc.get_jobs():
            logger = get_logger(job_id=job_id)
            logger.debug("Processing: done=%s", done)
            if done:
                backend.jobs_gc.remove_job_data(job_id)
                logger.info("Removed done job")
            else:
                backend.jobs_gc.push_back_job(job_id)
                logger.info("Pushed-back unfinished job")
            processed += 1
            self._processed += 1
            self._dump_collector_state(backend)
        return bool(processed)

    def _dump_collector_state(self, backend):
        self.dump_app_state(backend, {"processed": self._processed})



# TODO: Move to separate package


import socket

from contextlog import get_logger

import manhole


# =====
def start(port, reinstall_on_fork=False):
    manhole.logger = get_logger()
    manhole.Manhole.get_socket = staticmethod(_make_get_socket(port))
    manhole.ManholeConnectionThread.check_credentials = staticmethod(_check_credentials)
    manhole.install(patch_fork=reinstall_on_fork)


# =====
def _make_get_socket(port):
    def get_socket():
        for res in socket.getaddrinfo(None, port, socket.AF_UNSPEC, socket.SOCK_STREAM, 0, socket.AI_PASSIVE):
            (af, sock_type, proto, _, sa) = res
            sa = sa[:2]
            try:
                sock = manhole._ORIGINAL_SOCKET(af, sock_type, proto)  # pylint: disable=protected-access
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            except socket.error:
                sock = None
                continue

            try:
                sock.bind(sa)
                sock.listen(0)
                manhole.logger.critical("Manhole opened on *:%d", port)
                return sock
            except socket.error:
                manhole.logger.exception("Cannot bind/listen to %s", sa)

        msg = "Cannot create manhole on *:{}".format(port)
        manhole.logger.error(msg)
        raise RuntimeError(msg)
    return get_socket


def _check_credentials(client):
    # XXX: Access for all clients
    return manhole.get_peercred(client)



import os

from flask import request

from . import (
    valid_object_name,
    Resource,
    ApiError,
)


# =====
class CasPathResource(Resource):
    name = "View CAS data"
    methods = ("GET", "DELETE")
    dynamic = True
    docstring = """
        GET    -- Returns a dict with CAS user data:
                  # =====
                  {
                      "status":   "ok",
                      "message":  "<...>",
                      "result": {
                          "children": [
                              {"name": "<child>", "url": "http://api/url/to/child"},
                              ...
                          ],
                          "data": {
                              "stored": "<ISO-8601-like-time>",
                              "value": <any_type>,
                              "version": <int>|null,
                          }|null,
                      },
                  }
                  # =====

                  Possible GET errors:
                      404 -- Non-existant CAS path.

        DELETE -- Remove node without children.
                  Possible DELETE errors:
                      400 -- Node not empty (has children).
                      404 -- Non-existant CAS path.
    """

    def __init__(self, pool):
        self._pool = pool

    def process_request(self, path):  # pylint: disable=arguments-differ
        path = "/".join(map(valid_object_name, filter(None, path.split("/"))))
        path = "/" + path
        backend = self._pool.get_backend()
        try:
            if request.method == "GET":
                return self._request_get(backend, path)
            elif request.method == "DELETE":
                return self._request_delete(backend, path)
        finally:
            self._pool.retrieve_backend(backend)

    def _request_get(self, backend, path):
        children = backend.cas_storage.get_children(path)
        if children is None:
            raise ApiError(404, "Path not found")
        return ({
            "children": [
                {"name": child, "url": os.path.join(request.base_url, child)}
                for child in children
            ],
            "data": backend.cas_storage.get_raw(path),
        }, self.name)

    def _request_delete(self, backend, path):
        retval = backend.cas_storage.delete(path)
        if retval is None:
            raise ApiError(400, "Path has children")
        if not retval:
            raise ApiError(404, "Path not found")
        return (None, self.name)


class CasRootResource(CasPathResource):
    methods = ("GET",)
    dynamic = False

    def process_request(self):  # pylint: disable=arguments-differ
        return super().process_request("")
